/* 
 * Copyright (C) 2010 by Björn Biesenbach <bjoern@bjoern-b.de>
 *
 * This file is part of dockextd
 *
 * dockextd is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.

 * dockextd is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with dockextd.  If not, see <http://www.gnu.org/licenses/>.
 */
#include <stdlib.h>
#include <stdint.h>
#include <termios.h>
#include <fcntl.h>
#include <string.h>
#include <stdio.h>
#include <glib.h>
#include <time.h>

#include "serial.h"

#undef G_LOG_DOMAIN
#define G_LOG_DOMAIN "serial"

#define SERIAL_DEVICE "/dev/dockext"
#define CMD_CLEAR_SCREEN 8
#define CMD_BACKLIGHT 9
#define CMD_POSITION 10

static GIOChannel *channel;
static int fd;

static int serial_initiated = 0;

void (*callback_button1)() = NULL;
void (*callback_button2)() = NULL;
void (*callback_button3)() = NULL;
void (*callback_button4)() = NULL;
void (*callback_ready)() = NULL;

static gboolean serialReceive
(GIOChannel *channel, GIOCondition condition, gpointer data);

void serialSetCallback(int num, void *callback){
    switch(num)
    {
        case 1: callback_button1 = callback; break;
        case 2: callback_button2 = callback; break;
        case 3: callback_button3 = callback; break;
        case 4: callback_button4 = callback; break;
    }
}

void serialSetCallbackReady(void *callback){
    callback_ready = callback;
}

static void process_command(char cmd){
    switch(cmd)
    {
        case '1':     if(callback_button1) callback_button1(); break;
        case '2':     if(callback_button2) callback_button2(); break;
        case '3':     if(callback_button3) callback_button3(); break;
        case '4':     if(callback_button4) callback_button4(); break;
    }
}

gboolean serialTryInit(gpointer data){
    struct termios newtio;
    GError *error = NULL;

    if(serial_initiated)
        return FALSE;

    /* open the device */
    fd = open(SERIAL_DEVICE, O_RDWR | O_NOCTTY /*| O_NDELAY*/ | O_NONBLOCK );
    if (fd <0) 
    {
        return TRUE;
    }

    memset(&newtio, 0, sizeof(newtio)); /* clear struct for new port settings */
    newtio.c_cflag = B19200 | CS8 | CLOCAL | CREAD;
    newtio.c_iflag = (ICANON);
    newtio.c_oflag = 0;
    tcflush(fd, TCIFLUSH);
    if(tcsetattr(fd,TCSANOW,&newtio) < 0)
    {
        return TRUE;
    }
    
    GIOChannel *serial_device_chan = g_io_channel_unix_new(fd);
    guint serial_watch = g_io_add_watch(serial_device_chan, G_IO_IN | G_IO_ERR | G_IO_HUP,
        (GIOFunc)serialReceive, NULL);
    g_io_channel_set_encoding(serial_device_chan, NULL, &error);
    g_io_channel_unref(serial_device_chan);
    channel = serial_device_chan;
    serial_initiated = 1;

    if(callback_ready)
        callback_ready();
    
    return FALSE;
}

static gboolean serialReceive
(GIOChannel *channel, GIOCondition condition, gpointer data){
    gchar buf[2048];
    gsize bytes_read;
    GError *error = NULL;
    GIOStatus status;
    gint i;
    
    status = g_io_channel_read_chars(channel, buf, sizeof(buf), &bytes_read, &error);
    if( status != G_IO_STATUS_NORMAL && status != G_IO_STATUS_AGAIN)
    {
        if(error)
        {
            g_error_free(error);
        }
        serial_initiated = 0;
        g_io_channel_shutdown(channel, 0, NULL);
        close(fd);
        g_timeout_add_seconds(1, serialTryInit, NULL);
        return FALSE;
    }
    buf[bytes_read] = '\0';

    for(i=0; i < bytes_read; i++)
    {
        process_command(buf[i]);
    }
    return TRUE;
}

void serialSetPosition(uint8_t position){

    char position_cmd = CMD_POSITION;

    if(!serial_initiated)
        return;
    g_io_channel_write_chars(channel, &position_cmd, 1, NULL, NULL);
    g_io_channel_write_chars(channel, &position, 1, NULL, NULL);
    g_io_channel_flush(channel, NULL);
}

void serialInit(){
    g_timeout_add_seconds(1, serialTryInit, NULL);
}

void serialClearScreen(){
    char clearScreen = CMD_CLEAR_SCREEN;

    if(!serial_initiated)
        return;

    g_io_channel_write_chars(channel, &clearScreen, 1, NULL, NULL);
    g_io_channel_flush(channel, NULL);
}

void serialSetText(char *text, uint8_t line, gboolean centered){

    uint8_t space = 0;
    static char linebuf[2][17];
    static gboolean centerbuf[2];
    static int firstrun = 1;
    
    if(!serial_initiated)
        return;
    if(firstrun){
        linebuf[0][0] = '\0';
        linebuf[1][0] = '\0';
        firstrun = 0;
    }
    g_assert(strlen(text) < 17);
    
    strcpy(linebuf[line-1],text);
    centerbuf[line-1] = centered;
    
    serialClearScreen();
    serialSetPosition(centerbuf[0]*(16-strlen(linebuf[0]))/2);
    g_io_channel_write_chars(channel, linebuf[0], strlen(linebuf[0]), NULL, NULL);
    g_io_channel_flush(channel, NULL);
    serialSetPosition(0x40+centerbuf[1]*(16-strlen(linebuf[1]))/2);
    g_io_channel_write_chars(channel, linebuf[1], strlen(linebuf[1]), NULL, NULL);
    g_io_channel_flush(channel, NULL);
}

void serialSetBacklight(uint8_t value){
    
    char backlight = CMD_BACKLIGHT;
    if(!serial_initiated)
        return;
    g_io_channel_write_chars(channel, &backlight, 1, NULL, NULL);
    g_io_channel_write_chars(channel, &value, 1, NULL, NULL);
    g_io_channel_flush(channel, NULL);
}
